# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/11

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the offset.

"""
from __future__ import annotations

import unittest
from decimal import Decimal

import httpx
from flask import Flask

from test_site import db
from test_site.lib import JournalEntryLineItemData, JournalEntryCurrencyData, \
    JournalEntryData, BaseTestData
from testlib import NEXT_URI, Accounts, create_test_app, get_client, \
    match_journal_entry_detail

PREFIX: str = "/accounting/journal-entries"
"""The URL prefix for the journal entry management."""


class OffsetTestCase(unittest.TestCase):
    """The offset test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.app: Flask = create_test_app()

        with self.app.app_context():
            from accounting.models import JournalEntry, JournalEntryLineItem
            JournalEntry.query.delete()
            JournalEntryLineItem.query.delete()

        self.client, self.csrf_token = get_client(self.app, "editor")
        self.data: OffsetTestData = OffsetTestData(self.app, "editor")
        self.data.populate()

    def test_add_receivable_offset(self) -> None:
        """Tests to add the receivable offset.

        :return: None.
        """
        from accounting.models import Account, JournalEntry
        create_uri: str = f"{PREFIX}/create/receipt?next=%2F_next"
        store_uri: str = f"{PREFIX}/store/receipt"
        form: dict[str, str]
        old_amount: Decimal
        response: httpx.Response

        journal_entry_data: JournalEntryData = JournalEntryData(
            self.data.l_r_or3d.journal_entry.days, [JournalEntryCurrencyData(
                "USD",
                [],
                [JournalEntryLineItemData(
                    Accounts.RECEIVABLE,
                    self.data.l_r_or1d.description, "300",
                    original_line_item=self.data.l_r_or1d),
                 JournalEntryLineItemData(
                     Accounts.RECEIVABLE,
                     self.data.l_r_or1d.description, "100",
                     original_line_item=self.data.l_r_or1d),
                 JournalEntryLineItemData(
                     Accounts.RECEIVABLE,
                     self.data.l_r_or3d.description, "100",
                     original_line_item=self.data.l_r_or3d)])])

        # Non-existing original line item ID
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-original_line_item_id"] = "9999"
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # The same debit or credit
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-original_line_item_id"] \
            = str(self.data.l_p_or1c.id)
        form["currency-1-credit-1-account_code"] = self.data.l_p_or1c.account
        form["currency-1-credit-1-amount"] = "100"
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # The original line item does not need offset
        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            account.is_need_offset = False
            db.session.commit()
        response = self.client.post(
            store_uri,
            data=journal_entry_data.new_form(self.csrf_token, NEXT_URI))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)
        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            account.is_need_offset = True
            db.session.commit()

        # The original line item is also an offset
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-original_line_item_id"] \
            = str(self.data.l_p_of1d.id)
        form["currency-1-credit-1-account_code"] = self.data.l_p_of1d.account
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not the same currency
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-code"] = "EUR"
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not the same account
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-account_code"] = Accounts.NOTES_RECEIVABLE
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not exceeding net balance - partially offset
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-amount"] \
            = str(journal_entry_data.currencies[0].credit[0].amount
                  + Decimal("0.01"))
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not exceeding net balance - unmatched
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-3-amount"] \
            = str(journal_entry_data.currencies[0].credit[2].amount
                  + Decimal("0.01"))
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not before the original line items
        old_days = journal_entry_data.days
        journal_entry_data.days = old_days + 1
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)
        journal_entry_data.days = old_days

        # Success
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        journal_entry_id: int \
            = match_journal_entry_detail(response.headers["Location"])
        with self.app.app_context():
            journal_entry = db.session.get(JournalEntry, journal_entry_id)
            for offset in journal_entry.currencies[0].credit:
                self.assertIsNotNone(offset.original_line_item_id)

    def test_edit_receivable_offset(self) -> None:
        """Tests to edit the receivable offset.

        :return: None.
        """
        from accounting.models import Account
        journal_entry_data: JournalEntryData = self.data.j_r_of2
        edit_uri: str = f"{PREFIX}/{journal_entry_data.id}/edit?next=%2F_next"
        update_uri: str = f"{PREFIX}/{journal_entry_data.id}/update"
        form: dict[str, str]
        response: httpx.Response

        journal_entry_data.days = self.data.j_r_or2.days
        journal_entry_data.currencies[0].debit[0].amount = Decimal("600")
        journal_entry_data.currencies[0].credit[0].amount = Decimal("600")
        journal_entry_data.currencies[0].debit[2].amount = Decimal("600")
        journal_entry_data.currencies[0].credit[2].amount = Decimal("600")

        # Non-existing original line item ID
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-original_line_item_id"] = "9999"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # The same debit or credit
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-original_line_item_id"] \
            = str(self.data.l_p_or1c.id)
        form["currency-1-credit-1-account_code"] = self.data.l_p_or1c.account
        form["currency-1-debit-1-amount"] = "100"
        form["currency-1-credit-1-amount"] = "100"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # The original line item does not need offset
        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            account.is_need_offset = False
            db.session.commit()
        response = self.client.post(
            update_uri,
            data=journal_entry_data.update_form(self.csrf_token, NEXT_URI))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)
        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            account.is_need_offset = True
            db.session.commit()

        # The original line item is also an offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-original_line_item_id"] \
            = str(self.data.l_p_of1d.id)
        form["currency-1-credit-1-account_code"] = self.data.l_p_of1d.account
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not the same currency
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-code"] = "EUR"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not the same account
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-account_code"] = Accounts.NOTES_RECEIVABLE
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not exceeding net balance - partially offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-amount"] \
            = str(journal_entry_data.currencies[0].debit[0].amount
                  + Decimal("0.01"))
        form["currency-1-credit-1-amount"] \
            = str(journal_entry_data.currencies[0].credit[0].amount
                  + Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not exceeding net balance - unmatched
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-3-amount"] \
            = str(journal_entry_data.currencies[0].debit[2].amount
                  + Decimal("0.01"))
        form["currency-1-credit-3-amount"] \
            = str(journal_entry_data.currencies[0].credit[2].amount
                  + Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not before the original line items
        old_days: int = journal_entry_data.days
        journal_entry_data.days = old_days + 1
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)
        journal_entry_data.days = old_days

        # Success
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{journal_entry_data.id}?next=%2F_next")

    def test_edit_receivable_original_line_item(self) -> None:
        """Tests to edit the receivable original line item.

        :return: None.
        """
        from accounting.models import JournalEntry
        journal_entry_data: JournalEntryData = self.data.j_r_or1
        edit_uri: str = f"{PREFIX}/{journal_entry_data.id}/edit?next=%2F_next"
        update_uri: str = f"{PREFIX}/{journal_entry_data.id}/update"
        form: dict[str, str]
        response: httpx.Response

        journal_entry_data.days = self.data.j_r_of1.days
        journal_entry_data.currencies[0].debit[0].amount = Decimal("800")
        journal_entry_data.currencies[0].credit[0].amount = Decimal("800")
        journal_entry_data.currencies[0].debit[1].amount = Decimal("3.4")
        journal_entry_data.currencies[0].credit[1].amount = Decimal("3.4")

        # Not the same currency
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-code"] = "EUR"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not the same account
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-account_code"] = Accounts.NOTES_RECEIVABLE
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not less than offset total - partially offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-amount"] \
            = str(journal_entry_data.currencies[0].debit[0].amount
                  - Decimal("0.01"))
        form["currency-1-credit-1-amount"] \
            = str(journal_entry_data.currencies[0].credit[0].amount
                  - Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not less than offset total - fully offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-2-amount"] \
            = str(journal_entry_data.currencies[0].debit[1].amount
                  - Decimal("0.01"))
        form["currency-1-credit-2-amount"] \
            = str(journal_entry_data.currencies[0].credit[1].amount
                  - Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not after the offset items
        old_days: int = journal_entry_data.days
        journal_entry_data.days = old_days - 1
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)
        journal_entry_data.days = old_days

        # Not deleting matched original line items
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        del form["currency-1-debit-1-id"]
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Success
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{journal_entry_data.id}?next=%2F_next")

        # The original line item is always before the offset item, even when
        # they happen in the same day.
        with self.app.app_context():
            journal_entry_or: JournalEntry | None = db.session.get(
                JournalEntry, journal_entry_data.id)
            self.assertIsNotNone(journal_entry_or)
            journal_entry_of: JournalEntry | None = db.session.get(
                JournalEntry, self.data.j_r_of1.id)
            self.assertIsNotNone(journal_entry_of)
            self.assertEqual(journal_entry_or.date, journal_entry_of.date)
            self.assertLess(journal_entry_or.no, journal_entry_of.no)

    def test_add_payable_offset(self) -> None:
        """Tests to add the payable offset.

        :return: None.
        """
        from accounting.models import Account, JournalEntry
        create_uri: str = f"{PREFIX}/create/disbursement?next=%2F_next"
        store_uri: str = f"{PREFIX}/store/disbursement"
        form: dict[str, str]
        response: httpx.Response

        journal_entry_data: JournalEntryData = JournalEntryData(
            self.data.l_p_or3c.journal_entry.days, [JournalEntryCurrencyData(
                "USD",
                [JournalEntryLineItemData(
                    Accounts.PAYABLE,
                    self.data.l_p_or1c.description, "500",
                    original_line_item=self.data.l_p_or1c),
                 JournalEntryLineItemData(
                     Accounts.PAYABLE,
                     self.data.l_p_or1c.description, "300",
                     original_line_item=self.data.l_p_or1c),
                 JournalEntryLineItemData(
                     Accounts.PAYABLE,
                     self.data.l_p_or3c.description, "120",
                     original_line_item=self.data.l_p_or3c)],
                [])])

        # Non-existing original line item ID
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-original_line_item_id"] = "9999"
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # The same debit or credit
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-original_line_item_id"] \
            = str(self.data.l_r_or1d.id)
        form["currency-1-debit-1-account_code"] = self.data.l_r_or1d.account
        form["currency-1-debit-1-amount"] = "100"
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # The original line item does not need offset
        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            account.is_need_offset = False
            db.session.commit()
        response = self.client.post(
            store_uri,
            data=journal_entry_data.new_form(self.csrf_token, NEXT_URI))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)
        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            account.is_need_offset = True
            db.session.commit()

        # The original line item is also an offset
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-original_line_item_id"] \
            = str(self.data.l_r_of1c.id)
        form["currency-1-debit-1-account_code"] = self.data.l_r_of1c.account
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not the same currency
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-code"] = "EUR"
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not the same account
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-account_code"] = Accounts.NOTES_PAYABLE
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not exceeding net balance - partially offset
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-amount"] \
            = str(journal_entry_data.currencies[0].debit[0].amount
                  + Decimal("0.01"))
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not exceeding net balance - unmatched
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-3-amount"] \
            = str(journal_entry_data.currencies[0].debit[2].amount
                  + Decimal("0.01"))
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Not before the original line items
        old_days: int = journal_entry_data.days
        journal_entry_data.days = old_days + 1
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)
        journal_entry_data.days = old_days

        # Success
        form = journal_entry_data.new_form(self.csrf_token, NEXT_URI)
        response = self.client.post(store_uri, data=form)
        self.assertEqual(response.status_code, 302)
        journal_entry_id: int \
            = match_journal_entry_detail(response.headers["Location"])
        with self.app.app_context():
            journal_entry = db.session.get(JournalEntry, journal_entry_id)
            for offset in journal_entry.currencies[0].debit:
                self.assertIsNotNone(offset.original_line_item_id)

    def test_edit_payable_offset(self) -> None:
        """Tests to edit the payable offset.

        :return: None.
        """
        from accounting.models import Account, JournalEntry
        journal_entry_data: JournalEntryData = self.data.j_p_of2
        edit_uri: str = f"{PREFIX}/{journal_entry_data.id}/edit?next=%2F_next"
        update_uri: str = f"{PREFIX}/{journal_entry_data.id}/update"
        form: dict[str, str]
        response: httpx.Response

        journal_entry_data.days = self.data.j_p_or2.days
        journal_entry_data.currencies[0].debit[0].amount = Decimal("1100")
        journal_entry_data.currencies[0].credit[0].amount = Decimal("1100")
        journal_entry_data.currencies[0].debit[2].amount = Decimal("900")
        journal_entry_data.currencies[0].credit[2].amount = Decimal("900")

        # Non-existing original line item ID
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-original_line_item_id"] = "9999"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # The same debit or credit
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-original_line_item_id"] \
            = str(self.data.l_r_or1d.id)
        form["currency-1-debit-1-account_code"] = self.data.l_r_or1d.account
        form["currency-1-debit-1-amount"] = "100"
        form["currency-1-credit-1-amount"] = "100"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # The original line item does not need offset
        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            account.is_need_offset = False
            db.session.commit()
        response = self.client.post(
            update_uri,
            data=journal_entry_data.update_form(self.csrf_token, NEXT_URI))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)
        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            account.is_need_offset = True
            db.session.commit()

        # The original line item is also an offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-original_line_item_id"] \
            = str(self.data.l_r_of1c.id)
        form["currency-1-debit-1-account_code"] = self.data.l_r_of1c.account
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not the same currency
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-code"] = "EUR"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not the same account
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-account_code"] = Accounts.NOTES_PAYABLE
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not exceeding net balance - partially offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-amount"] \
            = str(journal_entry_data.currencies[0].debit[0].amount
                  + Decimal("0.01"))
        form["currency-1-credit-1-amount"] \
            = str(journal_entry_data.currencies[0].credit[0].amount
                  + Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not exceeding net balance - unmatched
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-3-amount"] \
            = str(journal_entry_data.currencies[0].debit[2].amount
                  + Decimal("0.01"))
        form["currency-1-credit-3-amount"] \
            = str(journal_entry_data.currencies[0].credit[2].amount
                  + Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not before the original line items
        old_days: int = journal_entry_data.days
        journal_entry_data.days = old_days + 1
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)
        journal_entry_data.days = old_days

        # Success
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        journal_entry_id: int \
            = match_journal_entry_detail(response.headers["Location"])
        with self.app.app_context():
            journal_entry = db.session.get(JournalEntry, journal_entry_id)
            for offset in journal_entry.currencies[0].debit:
                self.assertIsNotNone(offset.original_line_item_id)

    def test_edit_payable_original_line_item(self) -> None:
        """Tests to edit the payable original line item.

        :return: None.
        """
        from accounting.models import JournalEntry
        journal_entry_data: JournalEntryData = self.data.j_p_or1
        edit_uri: str = f"{PREFIX}/{journal_entry_data.id}/edit?next=%2F_next"
        update_uri: str = f"{PREFIX}/{journal_entry_data.id}/update"
        form: dict[str, str]
        response: httpx.Response

        journal_entry_data.days = self.data.j_p_of1.days
        journal_entry_data.currencies[0].debit[0].amount = Decimal("1200")
        journal_entry_data.currencies[0].credit[0].amount = Decimal("1200")
        journal_entry_data.currencies[0].debit[1].amount = Decimal("0.9")
        journal_entry_data.currencies[0].credit[1].amount = Decimal("0.9")

        # Not the same currency
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-code"] = "EUR"
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not the same account
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-credit-1-account_code"] = Accounts.NOTES_PAYABLE
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not less than offset total - partially offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-1-amount"] \
            = str(journal_entry_data.currencies[0].debit[0].amount
                  - Decimal("0.01"))
        form["currency-1-credit-1-amount"] \
            = str(journal_entry_data.currencies[0].credit[0].amount
                  - Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not less than offset total - fully offset
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        form["currency-1-debit-2-amount"] \
            = str(journal_entry_data.currencies[0].debit[1].amount
                  - Decimal("0.01"))
        form["currency-1-credit-2-amount"] \
            = str(journal_entry_data.currencies[0].credit[1].amount
                  - Decimal("0.01"))
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Not after the offset items
        old_days: int = journal_entry_data.days
        journal_entry_data.days = old_days - 1
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)
        journal_entry_data.days = old_days

        # Not deleting matched original line items
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        del form["currency-1-credit-1-id"]
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Success
        form = journal_entry_data.update_form(self.csrf_token, NEXT_URI)
        response = self.client.post(update_uri, data=form)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{journal_entry_data.id}?next=%2F_next")

        # The original line item is always before the offset item, even when
        # they happen in the same day
        with self.app.app_context():
            journal_entry_or: JournalEntry | None = db.session.get(
                JournalEntry, journal_entry_data.id)
            self.assertIsNotNone(journal_entry_or)
            journal_entry_of: JournalEntry | None = db.session.get(
                JournalEntry, self.data.j_p_of1.id)
            self.assertIsNotNone(journal_entry_of)
            self.assertEqual(journal_entry_or.date, journal_entry_of.date)
            self.assertLess(journal_entry_or.no, journal_entry_of.no)


class OffsetTestData(BaseTestData):
    """The offset test data."""

    def _init_data(self) -> None:
        # Receivable original line items
        self.l_r_or1d, self.l_r_or1c = self._couple(
            "Accountant", "1200", Accounts.RECEIVABLE, Accounts.SERVICE)
        self.l_r_or2d, self.l_r_or2c = self._couple(
            "Toy", "600", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or3d, self.l_r_or3c = self._couple(
            "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or4d, self.l_r_or4c = self._couple(
            "Interest", "3.4", Accounts.RECEIVABLE, Accounts.INTEREST)

        # Payable original line items
        self.l_p_or1d, self.l_p_or1c = self._couple(
            "Airplane", "2000", Accounts.TRAVEL, Accounts.PAYABLE)
        self.l_p_or2d, self.l_p_or2c = self._couple(
            "Phone", "900", Accounts.OFFICE, Accounts.PAYABLE)
        self.l_p_or3d, self.l_p_or3c = self._couple(
            "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or4d, self.l_p_or4c = self._couple(
            "Envelop", "0.9", Accounts.OFFICE, Accounts.PAYABLE)

        # Original journal entries
        self.j_r_or1: JournalEntryData = JournalEntryData(
            50, [JournalEntryCurrencyData(
                "USD", [self.l_r_or1d, self.l_r_or4d],
                [self.l_r_or1c, self.l_r_or4c])])
        self.j_r_or2: JournalEntryData = JournalEntryData(
            30, [JournalEntryCurrencyData(
                "USD", [self.l_r_or2d, self.l_r_or3d],
                [self.l_r_or2c, self.l_r_or3c])])
        self.j_p_or1: JournalEntryData = JournalEntryData(
            40, [JournalEntryCurrencyData(
                "USD", [self.l_p_or1d, self.l_p_or4d],
                [self.l_p_or1c, self.l_p_or4c])])
        self.j_p_or2: JournalEntryData = JournalEntryData(
            20, [JournalEntryCurrencyData(
                "USD", [self.l_p_or2d, self.l_p_or3d],
                [self.l_p_or2c, self.l_p_or3c])])

        self._add_journal_entry(self.j_r_or1)
        self._add_journal_entry(self.j_r_or2)
        self._add_journal_entry(self.j_p_or1)
        self._add_journal_entry(self.j_p_or2)

        # Receivable offset items
        self.l_r_of1d, self.l_r_of1c = self._couple(
            "Accountant", "500", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of1c.original_line_item = self.l_r_or1d
        self.l_r_of2d, self.l_r_of2c = self._couple(
            "Accountant", "200", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of2c.original_line_item = self.l_r_or1d
        self.l_r_of3d, self.l_r_of3c = self._couple(
            "Accountant", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of3c.original_line_item = self.l_r_or1d
        self.l_r_of4d, self.l_r_of4c = self._couple(
            "Toy", "240", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of4c.original_line_item = self.l_r_or2d
        self.l_r_of5d, self.l_r_of5c = self._couple(
            "Interest", "3.4", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of5c.original_line_item = self.l_r_or4d

        # Payable offset items
        self.l_p_of1d, self.l_p_of1c = self._couple(
            "Airplane", "800", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of1d.original_line_item = self.l_p_or1c
        self.l_p_of2d, self.l_p_of2c = self._couple(
            "Airplane", "300", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of2d.original_line_item = self.l_p_or1c
        self.l_p_of3d, self.l_p_of3c = self._couple(
            "Airplane", "100", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of3d.original_line_item = self.l_p_or1c
        self.l_p_of4d, self.l_p_of4c = self._couple(
            "Phone", "400", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of4d.original_line_item = self.l_p_or2c
        self.l_p_of5d, self.l_p_of5c = self._couple(
            "Envelop", "0.9", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of5d.original_line_item = self.l_p_or4c

        # Offset journal entries
        self.j_r_of1: JournalEntryData = JournalEntryData(
            25, [JournalEntryCurrencyData(
                "USD", [self.l_r_of1d], [self.l_r_of1c])])
        self.j_r_of2: JournalEntryData = JournalEntryData(
            20, [JournalEntryCurrencyData(
                "USD", [self.l_r_of2d, self.l_r_of3d, self.l_r_of4d],
                [self.l_r_of2c, self.l_r_of3c, self.l_r_of4c])])
        self.j_r_of3: JournalEntryData = JournalEntryData(
            15, [JournalEntryCurrencyData(
                "USD", [self.l_r_of5d], [self.l_r_of5c])])
        self.j_p_of1: JournalEntryData = JournalEntryData(
            15, [JournalEntryCurrencyData(
                "USD", [self.l_p_of1d], [self.l_p_of1c])])
        self.j_p_of2: JournalEntryData = JournalEntryData(
            10, [JournalEntryCurrencyData(
                "USD", [self.l_p_of2d, self.l_p_of3d, self.l_p_of4d],
                [self.l_p_of2c, self.l_p_of3c, self.l_p_of4c])])
        self.j_p_of3: JournalEntryData = JournalEntryData(
            5, [JournalEntryCurrencyData(
                "USD", [self.l_p_of5d], [self.l_p_of5c])])

        self._add_journal_entry(self.j_r_of1)
        self._add_journal_entry(self.j_r_of2)
        self._add_journal_entry(self.j_r_of3)
        self._add_journal_entry(self.j_p_of1)
        self._add_journal_entry(self.j_p_of2)
        self._add_journal_entry(self.j_p_of3)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The user utilities.

This module should not import any other module from the application.

"""
import typing as t
from abc import ABC, abstractmethod

import sqlalchemy as sa
from flask import g, Response
from flask_sqlalchemy.model import Model

T = t.TypeVar("T", bound=Model)


class UserUtilityInterface(t.Generic[T], ABC):
    """The interface for the user utilities."""

    @abstractmethod
    def can_view(self) -> bool:
        """Returns whether the currently logged-in user can view the accounting
        data.

        :return: True if the currently logged-in user can view the accounting
            data, or False otherwise.
        """

    @abstractmethod
    def can_edit(self) -> bool:
        """Returns whether the currently logged-in user can edit the accounting
        data.

        :return: True if the currently logged-in user can edit the accounting
            data, or False otherwise.
        """

    @abstractmethod
    def can_admin(self) -> bool:
        """Returns whether the currently logged-in user can administrate the
        accounting settings.

        :return: True if the currently logged-in user can administrate the
            accounting settings, or False otherwise.
        """

    @abstractmethod
    def unauthorized(self) -> Response | None:
        """Returns the response to require the user to log in.

        This may be a redirection to the login page, or an HTTP 401
        Unauthorized response for HTTP Authentication.  If this returns None,
        an HTTP 403 Forbidden response is return to the user.

        :return: The response to require the user to log in.
        """

    @property
    @abstractmethod
    def cls(self) -> t.Type[T]:
        """Returns the class of the user data model.

        :return: The class of the user data model.
        """

    @property
    @abstractmethod
    def pk_column(self) -> sa.Column:
        """Returns the primary key column.

        :return: The primary key column.
        """

    @property
    @abstractmethod
    def current_user(self) -> T | None:
        """Returns the currently logged-in user.

        :return: The currently logged-in user, or None if the user has not
            logged in
        """

    @abstractmethod
    def get_by_username(self, username: str) -> T | None:
        """Returns the user by her username.

        :return: The user by her username, or None if the user was not found.
        """

    @abstractmethod
    def get_pk(self, user: T) -> int:
        """Returns the primary key of the user, as an integer.

        :return: The primary key of the user, as an integer.
        """


__user_utils: UserUtilityInterface
"""The user utilities."""
user_cls: t.Type[Model] = Model
"""The user class."""
user_pk_column: sa.Column = sa.Column(sa.Integer)
"""The primary key column of the user class."""


def init_user_utils(utils: UserUtilityInterface) -> None:
    """Initializes the user utilities.

    :param utils: The user utilities.
    :return: None.
    """
    global __user_utils, user_cls, user_pk_column
    __user_utils = utils
    user_cls = utils.cls
    user_pk_column = utils.pk_column


def get_current_user_pk() -> int:
    """Returns the primary key value of the currently logged-in user.

    :return: The primary key value of the currently logged-in user.
    """
    return __user_utils.get_pk(get_current_user())


def has_user(username: str) -> bool:
    """Returns whether a user by the username exists.

    :param username: The username.
    :return: True if the user by the username exists, or False otherwise.
    """
    return __user_utils.get_by_username(username) is not None


def get_user_pk(username: str) -> int:
    """Returns the primary key value of the user by the username.

    :param username: The username.
    :return: The primary key value of the user by the username.
    """
    return __user_utils.get_pk(__user_utils.get_by_username(username))


def get_current_user() -> user_cls | None:
    """Returns the currently logged-in user.  The result is cached in the
    current request.

    :return: The currently logged-in user.
    """
    if not hasattr(g, "_accounting_user"):
        setattr(g, "_accounting_user", __user_utils.current_user)
    return getattr(g, "_accounting_user")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/8

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the unmatched offsets.

"""
import unittest

import httpx
from flask import Flask

from test_site import db
from test_site.lib import JournalEntryCurrencyData, JournalEntryData, \
    BaseTestData
from testlib import create_test_app, get_client, Accounts

PREFIX: str = "/accounting/unmatched-offsets"
"""The URL prefix for the unmatched offset management."""


class UnmatchedOffsetTestCase(unittest.TestCase):
    """The unmatched offset test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.app: Flask = create_test_app()

        with self.app.app_context():
            from accounting.models import JournalEntry, JournalEntryLineItem
            JournalEntry.query.delete()
            JournalEntryLineItem.query.delete()

        self.client, self.csrf_token = get_client(self.app, "editor")

    def test_nobody(self) -> None:
        """Test the permission as nobody.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "nobody")
        DifferentTestData(self.app, "nobody").populate()
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{Accounts.PAYABLE}",
                               data={"csrf_token": csrf_token})
        self.assertEqual(response.status_code, 403)

    def test_viewer(self) -> None:
        """Test the permission as viewer.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "viewer")
        DifferentTestData(self.app, "viewer").populate()
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{Accounts.PAYABLE}",
                               data={"csrf_token": csrf_token})
        self.assertEqual(response.status_code, 403)

    def test_editor(self) -> None:
        """Test the permission as editor.

        :return: None.
        """
        DifferentTestData(self.app, "editor").populate()
        response: httpx.Response

        response = self.client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(f"{PREFIX}/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(f"{PREFIX}/{Accounts.PAYABLE}",
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{Accounts.PAYABLE}")

    def test_empty_db(self) -> None:
        """Test the empty database.

        :return: None.
        """
        response: httpx.Response

        response = self.client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(f"{PREFIX}/{Accounts.PAYABLE}")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(f"{PREFIX}/{Accounts.PAYABLE}",
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"],
                         f"{PREFIX}/{Accounts.PAYABLE}")

    def test_different(self) -> None:
        """Tests to match against different descriptions and amounts.

        :return: None.
        """
        from accounting.models import Account, JournalEntryLineItem
        from accounting.utils.offset_matcher import OffsetMatcher
        data: DifferentTestData = DifferentTestData(self.app, "editor")
        data.populate()
        account: Account | None
        line_item: JournalEntryLineItem | None
        matcher: OffsetMatcher
        list_uri: str
        match_uri: str
        response: httpx.Response

        # The receivables
        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_r_or1d.id, data.l_r_or2d.id,
                              data.l_r_or3d.id, data.l_r_or4d.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_r_of1c.id, data.l_r_of2c.id,
                              data.l_r_of3c.id, data.l_r_of4c.id,
                              data.l_r_of5c.id})
            self.assertEqual({(x.original_line_item.id, x.offset.id)
                              for x in matcher.matched_pairs},
                             {(data.l_r_or4d.id, data.l_r_of5c.id)})
            for line_item_id in {data.l_r_of1c.id, data.l_r_of2c.id,
                                 data.l_r_of3c.id, data.l_r_of4c.id,
                                 data.l_r_of5c.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)

        list_uri = f"{PREFIX}/{Accounts.RECEIVABLE}"
        match_uri = f"{PREFIX}/{Accounts.RECEIVABLE}"
        response = self.client.post(match_uri,
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], list_uri)

        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_r_or1d.id, data.l_r_or2d.id,
                              data.l_r_or3d.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_r_of1c.id, data.l_r_of2c.id,
                              data.l_r_of3c.id, data.l_r_of4c.id})
            self.assertEqual(matcher.matches, 0)
            for line_item_id in {data.l_r_of1c.id, data.l_r_of2c.id,
                                 data.l_r_of3c.id, data.l_r_of4c.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)
            line_item = db.session.get(JournalEntryLineItem, data.l_r_of5c.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_r_or4d.id)

        # The payables
        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_p_or1c.id, data.l_p_or2c.id,
                              data.l_p_or3c.id, data.l_p_or4c.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_p_of1d.id, data.l_p_of2d.id,
                              data.l_p_of3d.id, data.l_p_of4d.id,
                              data.l_p_of5d.id})
            self.assertEqual({(x.original_line_item.id, x.offset.id)
                              for x in matcher.matched_pairs},
                             {(data.l_p_or4c.id, data.l_p_of5d.id)})
            for line_item_id in {data.l_p_of1d.id, data.l_p_of2d.id,
                                 data.l_p_of3d.id, data.l_p_of4d.id,
                                 data.l_p_of5d.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)

        list_uri = f"{PREFIX}/{Accounts.PAYABLE}"
        match_uri = f"{PREFIX}/{Accounts.PAYABLE}"
        response = self.client.post(match_uri,
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], list_uri)

        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_p_or1c.id, data.l_p_or2c.id,
                              data.l_p_or3c.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_p_of1d.id, data.l_p_of2d.id,
                              data.l_p_of3d.id, data.l_p_of4d.id})
            self.assertEqual(matcher.matches, 0)
            for line_item_id in {data.l_p_of1d.id, data.l_p_of2d.id,
                                 data.l_p_of3d.id, data.l_p_of4d.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)
            line_item = db.session.get(JournalEntryLineItem, data.l_p_of5d.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_p_or4c.id)

    def test_same(self) -> None:
        """Tests to match against same descriptions and amounts.

        :return: None.
        """
        from accounting.models import Account, JournalEntryLineItem
        from accounting.utils.offset_matcher import OffsetMatcher
        data: SameTestData = SameTestData(self.app, "editor")
        data.populate()
        account: Account | None
        line_item: JournalEntryLineItem | None
        matcher: OffsetMatcher
        list_uri: str
        match_uri: str
        response: httpx.Response

        # The receivables
        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_r_or1d.id, data.l_r_or3d.id,
                              data.l_r_or4d.id, data.l_r_or5d.id,
                              data.l_r_or6d.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_r_of1c.id, data.l_r_of2c.id,
                              data.l_r_of4c.id, data.l_r_of5c.id,
                              data.l_r_of6c.id})
            self.assertEqual({(x.original_line_item.id, x.offset.id)
                              for x in matcher.matched_pairs},
                             {(data.l_r_or1d.id, data.l_r_of2c.id),
                              (data.l_r_or3d.id, data.l_r_of4c.id),
                              (data.l_r_or4d.id, data.l_r_of6c.id)})
            for line_item_id in {data.l_r_of1c.id, data.l_r_of2c.id,
                                 data.l_r_of4c.id, data.l_r_of5c.id,
                                 data.l_r_of6c.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)
            line_item = db.session.get(JournalEntryLineItem, data.l_r_of3c.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_r_or2d.id)

        list_uri = f"{PREFIX}/{Accounts.RECEIVABLE}"
        match_uri = f"{PREFIX}/{Accounts.RECEIVABLE}"
        response = self.client.post(match_uri,
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], list_uri)

        with self.app.app_context():
            account = Account.find_by_code(Accounts.RECEIVABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_r_or5d.id, data.l_r_or6d.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_r_of1c.id, data.l_r_of5c.id})
            self.assertEqual(matcher.matches, 0)
            for line_item_id in {data.l_r_of1c.id, data.l_r_of5c.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)
            line_item = db.session.get(JournalEntryLineItem, data.l_r_of2c.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_r_or1d.id)
            line_item = db.session.get(JournalEntryLineItem, data.l_r_of3c.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_r_or2d.id)
            line_item = db.session.get(JournalEntryLineItem, data.l_r_of4c.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_r_or3d.id)
            line_item = db.session.get(JournalEntryLineItem, data.l_r_of6c.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_r_or4d.id)

        # The payables
        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_p_or1c.id, data.l_p_or3c.id,
                              data.l_p_or4c.id, data.l_p_or5c.id,
                              data.l_p_or6c.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_p_of1d.id, data.l_p_of2d.id,
                              data.l_p_of4d.id, data.l_p_of5d.id,
                              data.l_p_of6d.id})
            self.assertEqual({(x.original_line_item.id, x.offset.id)
                              for x in matcher.matched_pairs},
                             {(data.l_p_or1c.id, data.l_p_of2d.id),
                              (data.l_p_or3c.id, data.l_p_of4d.id),
                              (data.l_p_or4c.id, data.l_p_of6d.id)})
            for line_item_id in {data.l_p_of1d.id, data.l_p_of2d.id,
                                 data.l_p_of4d.id, data.l_p_of5d.id,
                                 data.l_p_of6d.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)
            line_item = db.session.get(JournalEntryLineItem, data.l_p_of3d.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_p_or2c.id)

        list_uri = f"{PREFIX}/{Accounts.PAYABLE}"
        match_uri = f"{PREFIX}/{Accounts.PAYABLE}"
        response = self.client.post(match_uri,
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], list_uri)

        with self.app.app_context():
            account = Account.find_by_code(Accounts.PAYABLE)
            assert account is not None
            matcher = OffsetMatcher(account)
            self.assertEqual({x.id for x in matcher.unapplied},
                             {data.l_p_or5c.id, data.l_p_or6c.id})
            self.assertEqual({x.id for x in matcher.unmatched_offsets},
                             {data.l_p_of1d.id, data.l_p_of5d.id})
            self.assertEqual(matcher.matches, 0)
            for line_item_id in {data.l_p_of1d.id, data.l_p_of5d.id}:
                line_item = db.session.get(JournalEntryLineItem, line_item_id)
                self.assertIsNotNone(line_item)
                self.assertIsNone(line_item.original_line_item_id)
            line_item = db.session.get(JournalEntryLineItem, data.l_p_of2d.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_p_or1c.id)
            line_item = db.session.get(JournalEntryLineItem, data.l_p_of3d.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_p_or2c.id)
            line_item = db.session.get(JournalEntryLineItem, data.l_p_of4d.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_p_or3c.id)
            line_item = db.session.get(JournalEntryLineItem, data.l_p_of6d.id)
            self.assertIsNotNone(line_item)
            self.assertIsNotNone(line_item.original_line_item_id)
            self.assertEqual(line_item.original_line_item_id, data.l_p_or4c.id)


class DifferentTestData(BaseTestData):
    """The test data for different descriptions and amounts."""

    def _init_data(self) -> None:
        # Receivable original line items
        self.l_r_or1d, self.l_r_or1c = self._couple(
            "Accountant", "1200", Accounts.RECEIVABLE, Accounts.SERVICE)
        self.l_r_or2d, self.l_r_or2c = self._couple(
            "Toy", "600", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or3d, self.l_r_or3c = self._couple(
            "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or4d, self.l_r_or4c = self._couple(
            "Interest", "3.4", Accounts.RECEIVABLE, Accounts.INTEREST)

        # Payable original line items
        self.l_p_or1d, self.l_p_or1c = self._couple(
            "Airplane", "2000", Accounts.TRAVEL, Accounts.PAYABLE)
        self.l_p_or2d, self.l_p_or2c = self._couple(
            "Phone", "900", Accounts.OFFICE, Accounts.PAYABLE)
        self.l_p_or3d, self.l_p_or3c = self._couple(
            "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or4d, self.l_p_or4c = self._couple(
            "Envelop", "0.9", Accounts.OFFICE, Accounts.PAYABLE)

        # Original journal entries
        self.j_r_or1: JournalEntryData = JournalEntryData(
            50, [JournalEntryCurrencyData(
                "USD", [self.l_r_or1d, self.l_r_or4d],
                [self.l_r_or1c, self.l_r_or4c])])
        self.j_r_or2: JournalEntryData = JournalEntryData(
            30, [JournalEntryCurrencyData(
                "USD", [self.l_r_or2d, self.l_r_or3d],
                [self.l_r_or2c, self.l_r_or3c])])
        self.j_p_or1: JournalEntryData = JournalEntryData(
            40, [JournalEntryCurrencyData(
                "USD", [self.l_p_or1d, self.l_p_or4d],
                [self.l_p_or1c, self.l_p_or4c])])
        self.j_p_or2: JournalEntryData = JournalEntryData(
            20, [JournalEntryCurrencyData(
                "USD", [self.l_p_or2d, self.l_p_or3d],
                [self.l_p_or2c, self.l_p_or3c])])

        self._add_journal_entry(self.j_r_or1)
        self._add_journal_entry(self.j_r_or2)
        self._add_journal_entry(self.j_p_or1)
        self._add_journal_entry(self.j_p_or2)

        # Receivable offset items
        self.l_r_of1d, self.l_r_of1c = self._couple(
            "Accountant", "500", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of2d, self.l_r_of2c = self._couple(
            "Accountant", "200", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of3d, self.l_r_of3c = self._couple(
            "Accountant", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of4d, self.l_r_of4c = self._couple(
            "Toy", "240", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of5d, self.l_r_of5c = self._couple(
            "Interest", "3.4", Accounts.CASH, Accounts.RECEIVABLE)

        # Payable offset items
        self.l_p_of1d, self.l_p_of1c = self._couple(
            "Airplane", "800", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of2d, self.l_p_of2c = self._couple(
            "Airplane", "300", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of3d, self.l_p_of3c = self._couple(
            "Airplane", "100", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of4d, self.l_p_of4c = self._couple(
            "Phone", "400", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of5d, self.l_p_of5c = self._couple(
            "Envelop", "0.9", Accounts.PAYABLE, Accounts.CASH)

        # Offset journal entries
        self.j_r_of1: JournalEntryData = JournalEntryData(
            25, [JournalEntryCurrencyData(
                "USD", [self.l_r_of1d], [self.l_r_of1c])])
        self.j_r_of2: JournalEntryData = JournalEntryData(
            20, [JournalEntryCurrencyData(
                "USD", [self.l_r_of2d, self.l_r_of3d, self.l_r_of4d],
                [self.l_r_of2c, self.l_r_of3c, self.l_r_of4c])])
        self.j_r_of3: JournalEntryData = JournalEntryData(
            15, [JournalEntryCurrencyData(
                "USD", [self.l_r_of5d], [self.l_r_of5c])])
        self.j_p_of1: JournalEntryData = JournalEntryData(
            15, [JournalEntryCurrencyData(
                "USD", [self.l_p_of1d], [self.l_p_of1c])])
        self.j_p_of2: JournalEntryData = JournalEntryData(
            10, [JournalEntryCurrencyData(
                "USD", [self.l_p_of2d, self.l_p_of3d, self.l_p_of4d],
                [self.l_p_of2c, self.l_p_of3c, self.l_p_of4c])])
        self.j_p_of3: JournalEntryData = JournalEntryData(
            5, [JournalEntryCurrencyData(
                "USD", [self.l_p_of5d], [self.l_p_of5c])])

        self._add_journal_entry(self.j_r_of1)
        self._add_journal_entry(self.j_r_of2)
        self._add_journal_entry(self.j_r_of3)
        self._add_journal_entry(self.j_p_of1)
        self._add_journal_entry(self.j_p_of2)
        self._add_journal_entry(self.j_p_of3)


class SameTestData(BaseTestData):
    """The test data with same descriptions and amounts."""

    def _init_data(self) -> None:
        # Receivable original line items
        self.l_r_or1d, self.l_r_or1c = self._add_simple_journal_entry(
            60, "USD", "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or2d, self.l_r_or2c = self._add_simple_journal_entry(
            50, "USD", "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or3d, self.l_r_or3c = self._add_simple_journal_entry(
            40, "USD", "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or4d, self.l_r_or4c = self._add_simple_journal_entry(
            30, "USD", "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or5d, self.l_r_or5c = self._add_simple_journal_entry(
            20, "USD", "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)
        self.l_r_or6d, self.l_r_or6c = self._add_simple_journal_entry(
            10, "USD", "Noodles", "100", Accounts.RECEIVABLE, Accounts.SALES)

        # Payable original line items
        self.l_p_or1d, self.l_p_or1c = self._add_simple_journal_entry(
            60, "USD", "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or2d, self.l_p_or2c = self._add_simple_journal_entry(
            50, "USD", "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or3d, self.l_p_or3c = self._add_simple_journal_entry(
            40, "USD", "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or4d, self.l_p_or4c = self._add_simple_journal_entry(
            30, "USD", "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or5d, self.l_p_or5c = self._add_simple_journal_entry(
            20, "USD", "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)
        self.l_p_or6d, self.l_p_or6c = self._add_simple_journal_entry(
            10, "USD", "Steak", "120", Accounts.MEAL, Accounts.PAYABLE)

        # Receivable offset items
        self.l_r_of1d, self.l_r_of1c = self._add_simple_journal_entry(
            65, "USD", "Noodles", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of2d, self.l_r_of2c = self._add_simple_journal_entry(
            35, "USD", "Noodles", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of3d, self.l_r_of3c = self._couple(
            "Noodles", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of3c.original_line_item = self.l_r_or2d
        j_r_of3: JournalEntryData = JournalEntryData(
            35, [JournalEntryCurrencyData(
                "USD", [self.l_r_of3d], [self.l_r_of3c])])
        self.l_r_of4d, self.l_r_of4c = self._add_simple_journal_entry(
            35, "USD", "Noodles", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of5d, self.l_r_of5c = self._add_simple_journal_entry(
            35, "USD", "Noodles", "100", Accounts.CASH, Accounts.RECEIVABLE)
        self.l_r_of6d, self.l_r_of6c = self._add_simple_journal_entry(
            15, "USD", "Noodles", "100", Accounts.CASH, Accounts.RECEIVABLE)

        # Payable offset items
        self.l_p_of1d, self.l_p_of1c = self._add_simple_journal_entry(
            65, "USD", "Steak", "120", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of2d, self.l_p_of2c = self._add_simple_journal_entry(
            35, "USD", "Steak", "120", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of3d, self.l_p_of3c = self._couple(
            "Steak", "120", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of3d.original_line_item = self.l_p_or2c
        j_p_of3: JournalEntryData = JournalEntryData(
            35, [JournalEntryCurrencyData(
                "USD", [self.l_p_of3d], [self.l_p_of3c])])
        self.l_p_of4d, self.l_p_of4c = self._add_simple_journal_entry(
            35, "USD", "Steak", "120", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of5d, self.l_p_of5c = self._add_simple_journal_entry(
            35, "USD", "Steak", "120", Accounts.PAYABLE, Accounts.CASH)
        self.l_p_of6d, self.l_p_of6c = self._add_simple_journal_entry(
            15, "USD", "Steak", "120", Accounts.PAYABLE, Accounts.CASH)

        self._add_journal_entry(j_r_of3)
        self._add_journal_entry(j_p_of3)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The template filters for the journal entry management.

"""
from decimal import Decimal
from html import escape
from urllib.parse import ParseResult, urlparse, parse_qsl, urlencode, \
    urlunparse

from flask import request


def with_type(uri: str) -> str:
    """Adds the journal entry type to the URI, if it is specified.

    :param uri: The URI.
    :return: The result URL, optionally with the journal entry type added.
    """
    if "as" not in request.args:
        return uri
    uri_p: ParseResult = urlparse(uri)
    params: list[tuple[str, str]] = parse_qsl(uri_p.query)
    params = [x for x in params if x[0] != "as"]
    params.append(("as", request.args["as"]))
    parts: list[str] = list(uri_p)
    parts[4] = urlencode(params)
    return urlunparse(parts)


def to_transfer(uri: str) -> str:
    """Adds the transfer journal entry type to the URI.

    :param uri: The URI.
    :return: The result URL, with the transfer journal entry type added.
    """
    uri_p: ParseResult = urlparse(uri)
    params: list[tuple[str, str]] = parse_qsl(uri_p.query)
    params = [x for x in params if x[0] != "as"]
    params.append(("as", "transfer"))
    parts: list[str] = list(uri_p)
    parts[4] = urlencode(params)
    return urlunparse(parts)


def format_amount_input(value: Decimal | None) -> str:
    """Format an amount for an input value.

    :param value: The amount.
    :return: The formatted amount text for an input value.
    """
    if value is None:
        return ""
    whole: int = int(value)
    frac: Decimal = (value - whole).normalize()
    return str(whole) + str(frac)[1:]


def text2html(value: str) -> str:
    """Converts plain text into HTML.

    :param value: The plain text.
    :return: The HTML.
    """
    s: str = escape(value)
    s = s.replace("\n", "<br>")
    s = s.replace("  ", " &nbsp;")
    return s



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/30

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The random ID mixin for the data models.

This module should not import any other module from the application.

"""
import typing as t
from secrets import randbelow

from accounting import db


def new_id(cls: t.Type):
    """Returns a new random ID for the data model.

    :param cls: The data model.
    :return: The generated new random ID.
    """
    while True:
        obj_id: int = 100000000 + randbelow(900000000)
        if db.session.get(cls, obj_id) is None:
            return obj_id



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/22

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The current assets and liabilities account.

"""
import typing as t

from accounting import db
from accounting.locale import gettext
from accounting.models import Account
import sqlalchemy as sa


class CurrentAccount:
    """A current assets and liabilities account."""
    CURRENT_AL_CODE: str = "0000-000"
    """The account code for all current assets and liabilities."""

    def __init__(self, account: Account | None = None):
        """Constructs the current assets and liabilities account.

        :param account: The actual account.
        """
        self.account: Account | None = account
        """The actual account."""
        self.id: int = -1 if account is None else account.id
        """The ID."""
        self.code: str = "" if account is None else account.code
        """The code."""
        self.title: str = "" if account is None else account.title
        """The title."""
        self.str: str = "" if account is None else str(account)
        """The string representation of the account."""

    def __str__(self) -> str:
        """Returns the string representation of the account.

        :return: The string representation of the account.
        """
        return self.str

    @classmethod
    def current_assets_and_liabilities(cls) -> t.Self:
        """Returns the pseudo account for all current assets and liabilities.

        :return: The pseudo account for all current assets and liabilities.
        """
        account: cls = cls()
        account.id = 0
        account.code = cls.CURRENT_AL_CODE
        account.title = gettext("current assets and liabilities")
        account.str = account.title
        return account

    @classmethod
    def accounts(cls) -> list[t.Self]:
        """Returns the current assets and liabilities accounts.

        :return: The current assets and liabilities accounts.
        """
        accounts: list[cls] = [cls.current_assets_and_liabilities()]
        accounts.extend([CurrentAccount(x)
                         for x in db.session.query(Account)
                        .filter(cls.sql_condition())
                        .order_by(Account.base_code, Account.no)])
        return accounts

    @classmethod
    def sql_condition(cls) -> sa.BinaryExpression:
        """Returns the SQL condition for the current assets and liabilities
        accounts.

        :return: The SQL condition for the current assets and liabilities
            accounts.
        """
        return sa.or_(Account.base_code.startswith("11"),
                      Account.base_code.startswith("12"),
                      Account.base_code.startswith("21"),
                      Account.base_code.startswith("22"))



#! env python3
# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/28

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The translation management utilities for the test site.

"""
import os
import re
from pathlib import Path
from time import strftime

import click
from babel.messages.frontend import CommandLineInterface
from opencc import OpenCC

root_dir: Path = Path(__file__).parent.parent
translation_dir: Path = root_dir / "tests" / "test_site" / "translations"
domain: str = "messages"


@click.group()
def main() -> None:
    """Manages the message translation."""


@click.command("extract")
def babel_extract() -> None:
    """Extracts the messages for translation."""
    os.chdir(root_dir)
    cfg: Path = translation_dir / "babel.cfg"
    pot: Path = translation_dir / f"{domain}.pot"
    zh_hant: Path = translation_dir / "zh_Hant" / "LC_MESSAGES"\
        / f"{domain}.po"
    zh_hans: Path = translation_dir / "zh_Hans" / "LC_MESSAGES"\
        / f"{domain}.po"
    CommandLineInterface().run([
        "pybabel", "extract", "-F", str(cfg), "-k", "lazy_gettext", "-k", "A_",
        "-o", str(pot), str(Path("tests") / "test_site")])
    if not zh_hant.exists():
        zh_hant.touch()
    if not zh_hans.exists():
        zh_hans.touch()
    CommandLineInterface().run([
        "pybabel", "update", "-i", str(pot), "-D", domain,
        "-d", translation_dir])


@click.command("compile")
def babel_compile() -> None:
    """Compiles the translated messages."""
    __convert_chinese()
    __update_rev_date()
    CommandLineInterface().run([
        "pybabel", "compile", "-D", domain, "-d", translation_dir])


def __convert_chinese() -> None:
    """Updates the Simplified Chinese translation according to the Traditional
    Chinese translation.

    :return: None.
    """
    cc: OpenCC = OpenCC("tw2sp")
    zh_hant: Path = translation_dir / "zh_Hant" / "LC_MESSAGES"\
        / f"{domain}.po"
    zh_hans: Path = translation_dir / "zh_Hans" / "LC_MESSAGES"\
        / f"{domain}.po"
    now: str = strftime("%Y-%m-%d %H:%M%z")
    with open(zh_hant, "r") as f:
        content: str = f.read()
    content = cc.convert(content)
    content = re.sub(r"^# Chinese \\(Traditional\\) translations ",
                     "# Chinese (Simplified) translations ", content)
    content = re.sub(r"\n\"PO-Revision-Date: [^\n]*\"\n",
                     f"\n\"PO-Revision-Date: {now}\\\\n\"\n",
                     content)
    content = content.replace("\n\"Language-Team: zh_Hant",
                              "\n\"Language-Team: zh_Hans")
    content = content.replace("\n\"Language: zh_Hant\\n\"\n",
                              "\n\"Language: zh_Hans\\n\"\n")
    content = content.replace("\nmsgstr \"zh-Hant\"\n",
                              "\nmsgstr \"zh-Hans\"\n")
    zh_hans.parent.mkdir(exist_ok=True)
    with open(zh_hans, "w") as f:
        f.write(content)


def __update_rev_date() -> None:
    """Updates the revision dates in the PO files.

    :return: None.
    """
    for language_dir in translation_dir.glob("*"):
        po_file: Path = language_dir / "LC_MESSAGES" / f"{domain}.po"
        if po_file.is_file():
            __update_file_rev_date(po_file)


def __update_file_rev_date(file: Path) -> None:
    """Updates the revision date of a PO file

    :param file: The PO file.
    :return: None.
    """
    now = strftime("%Y-%m-%d %H:%M%z")
    with open(file, "r+") as f:
        content = f.read()
        content = re.sub(r"\n\"PO-Revision-Date: [^\n]*\"\n",
                         f"\n\"PO-Revision-Date: {now}\\\\n\"\n",
                         content)
        f.seek(0)
        f.write(content)


main.add_command(babel_extract)
main.add_command(babel_compile)

if __name__ == "__main__":
    main()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/19

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The path converters for the journal entry management.

"""
from datetime import date

from flask import abort
from werkzeug.routing import BaseConverter

from accounting import db
from accounting.models import JournalEntry
from accounting.utils.journal_entry_types import JournalEntryType


class JournalEntryConverter(BaseConverter):
    """The journal entry converter to convert the journal entry ID from and to
    the corresponding journal entry in the routes."""

    def to_python(self, value: str) -> JournalEntry:
        """Converts a journal entry ID to a journal entry.

        :param value: The journal entry ID.
        :return: The corresponding journal entry.
        """
        journal_entry: JournalEntry | None = db.session.get(JournalEntry, value)
        if journal_entry is None:
            abort(404)
        return journal_entry

    def to_url(self, value: JournalEntry) -> str:
        """Converts a journal entry to its ID.

        :param value: The journal entry.
        :return: The ID.
        """
        return str(value.id)


class JournalEntryTypeConverter(BaseConverter):
    """The journal entry converter to convert the journal entry type ID from
    and to the corresponding journal entry type in the routes."""

    def to_python(self, value: str) -> JournalEntryType:
        """Converts a journal entry ID to a journal entry.

        :param value: The journal entry ID.
        :return: The corresponding journal entry type.
        """
        type_dict: dict[str, JournalEntryType] \
            = {x.value: x for x in JournalEntryType}
        journal_entry_type: JournalEntryType | None = type_dict.get(value)
        if journal_entry_type is None:
            abort(404)
        return journal_entry_type

    def to_url(self, value: JournalEntryType) -> str:
        """Converts a journal entry type to its ID.

        :param value: The journal entry type.
        :return: The ID.
        """
        return str(value.value)


class DateConverter(BaseConverter):
    """The date converter to convert the ISO date from and to the
    corresponding date in the routes."""

    def to_python(self, value: str) -> date:
        """Converts an ISO date to a date.

        :param value: The ISO date.
        :return: The corresponding date.
        """
        try:
            return date.fromisoformat(value)
        except ValueError:
            abort(404)

    def to_url(self, value: date) -> str:
        """Converts a date to its ISO date.

        :param value: The date.
        :return: The ISO date.
        """
        return value.isoformat()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/15

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utility to cast a SQLAlchemy column into the column type, to avoid
warnings from the IDE.

This module should not import any other module from the application.

"""
import typing as t

import sqlalchemy as sa


def be(expression: t.Any) -> sa.BinaryExpression:
    """Casts the SQLAlchemy binary expression to the binary expression type.

    :param expression: The binary expression.
    :return: The binary expression itself.
    """
    assert isinstance(expression, sa.BinaryExpression)
    return expression


def s(message: t.Any) -> str:
    """Casts the LazyString message to the string type.

    :param message: The message.
    :return: The binary expression itself.
    """
    return message



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/6

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The path converters for the currency management.

"""
from flask import abort
from werkzeug.routing import BaseConverter

from accounting import db
from accounting.models import Currency


class CurrencyConverter(BaseConverter):
    """The currency converter to convert the currency code and to the
    corresponding currency in the routes."""

    def to_python(self, value: str) -> Currency:
        """Converts a currency code to a currency.

        :param value: The currency code.
        :return: The corresponding currency.
        """
        currency: Currency | None = db.session.get(Currency, value)
        if currency is None:
            abort(404)
        return currency

    def to_url(self, value: Currency) -> str:
        """Converts a currency to its code.

        :param value: The currency.
        :return: The code.
        """
        return value.code



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/1

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the currency management.

"""
import unittest
from datetime import timedelta, date

import httpx
from flask import Flask

from test_site import db
from testlib import NEXT_URI, create_test_app, get_client, set_locale, \
    add_journal_entry


class CurrencyData:
    """The currency data."""

    def __init__(self, code: str, name: str):
        """Constructs the currency data.

        :param code: The code.
        :param name: The name.
        """
        self.code: str = code
        """The code."""
        self.name: str = name
        """The name."""


USD: CurrencyData = CurrencyData("USD", "US Dollar")
"""The US dollars."""
EUR: CurrencyData = CurrencyData("EUR", "Euro")
"""The European dollars."""
TWD: CurrencyData = CurrencyData("TWD", "Taiwan dollars")
"""The Taiwan dollars."""
JPY: CurrencyData = CurrencyData("JPY", "Japanese yen")
"""The Japanese yen."""
PREFIX: str = "/accounting/currencies"
"""The URL prefix for the currency management."""


class CurrencyTestCase(unittest.TestCase):
    """The currency test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.app: Flask = create_test_app()

        with self.app.app_context():
            from accounting.models import Currency, CurrencyL10n
            CurrencyL10n.query.delete()
            Currency.query.delete()
            db.session.commit()

        self.client, self.csrf_token = get_client(self.app, "editor")
        response: httpx.Response

        response = self.client.post(f"{PREFIX}/store",
                                    data={"csrf_token": self.csrf_token,
                                          "code": USD.code,
                                          "name": USD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{USD.code}")

        response = self.client.post(f"{PREFIX}/store",
                                    data={"csrf_token": self.csrf_token,
                                          "code": EUR.code,
                                          "name": EUR.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{EUR.code}")

    def test_nobody(self) -> None:
        """Test the permission as nobody.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "nobody")
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{USD.code}")
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/create")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/store",
                               data={"csrf_token": csrf_token,
                                     "code": TWD.code,
                                     "name": TWD.name})
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{USD.code}/edit")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{USD.code}/update",
                               data={"csrf_token": csrf_token,
                                     "code": JPY.code,
                                     "name": JPY.name})
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{EUR.code}/delete",
                               data={"csrf_token": csrf_token})
        self.assertEqual(response.status_code, 403)

    def test_viewer(self) -> None:
        """Test the permission as viewer.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "viewer")
        response: httpx.Response

        response = client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/{USD.code}")
        self.assertEqual(response.status_code, 200)

        response = client.get(f"{PREFIX}/create")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/store",
                               data={"csrf_token": csrf_token,
                                     "code": TWD.code,
                                     "name": TWD.name})
        self.assertEqual(response.status_code, 403)

        response = client.get(f"{PREFIX}/{USD.code}/edit")
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{USD.code}/update",
                               data={"csrf_token": csrf_token,
                                     "code": JPY.code,
                                     "name": JPY.name})
        self.assertEqual(response.status_code, 403)

        response = client.post(f"{PREFIX}/{EUR.code}/delete",
                               data={"csrf_token": csrf_token})
        self.assertEqual(response.status_code, 403)

    def test_editor(self) -> None:
        """Test the permission as editor.

        :return: None.
        """
        response: httpx.Response

        response = self.client.get(PREFIX)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(f"{PREFIX}/{USD.code}")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(f"{PREFIX}/create")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(f"{PREFIX}/store",
                                    data={"csrf_token": self.csrf_token,
                                          "code": TWD.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{TWD.code}")

        response = self.client.get(f"{PREFIX}/{USD.code}/edit")
        self.assertEqual(response.status_code, 200)

        response = self.client.post(f"{PREFIX}/{USD.code}/update",
                                    data={"csrf_token": self.csrf_token,
                                          "code": JPY.code,
                                          "name": JPY.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{JPY.code}")

        response = self.client.post(f"{PREFIX}/{EUR.code}/delete",
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], PREFIX)

    def test_add(self) -> None:
        """Tests to add the currencies.

        :return: None.
        """
        from accounting.models import Currency
        create_uri: str = f"{PREFIX}/create"
        store_uri: str = f"{PREFIX}/store"
        detail_uri: str = f"{PREFIX}/{TWD.code}"
        response: httpx.Response

        with self.app.app_context():
            self.assertEqual({x.code for x in Currency.query.all()},
                             {USD.code, EUR.code})

        # Missing CSRF token
        response = self.client.post(store_uri,
                                    data={"code": TWD.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 400)

        # CSRF token mismatch
        response = self.client.post(store_uri,
                                    data={"csrf_token": f"{self.csrf_token}-2",
                                          "code": TWD.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 400)

        # Empty code
        response = self.client.post(store_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": " ",
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Blocked code, with spaces to be stripped
        response = self.client.post(store_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": " create ",
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Bad code
        response = self.client.post(store_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": " zzc ",
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Empty name
        response = self.client.post(store_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": TWD.code,
                                          "name": " "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        # Success, with spaces to be stripped
        response = self.client.post(store_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": f" {TWD.code} ",
                                          "name": f" {TWD.name} "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        # Duplicated code
        response = self.client.post(store_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": TWD.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], create_uri)

        with self.app.app_context():
            self.assertEqual({x.code for x in Currency.query.all()},
                             {USD.code, EUR.code, TWD.code})

            currency: Currency = db.session.get(Currency, TWD.code)
            self.assertEqual(currency.code, TWD.code)
            self.assertEqual(currency.name_l10n, TWD.name)

    def test_basic_update(self) -> None:
        """Tests the basic rules to update a user.

        :return: None.
        """
        from accounting.models import Currency
        detail_uri: str = f"{PREFIX}/{USD.code}"
        edit_uri: str = f"{PREFIX}/{USD.code}/edit"
        update_uri: str = f"{PREFIX}/{USD.code}/update"
        detail_c_uri: str = f"{PREFIX}/{TWD.code}"
        response: httpx.Response

        # Success, with spaces to be stripped
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": f" {USD.code} ",
                                          "name": f" {USD.name}-1 "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency: Currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.code, USD.code)
            self.assertEqual(currency.name_l10n, f"{USD.name}-1")

        # Empty code
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": " ",
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Blocked code, with spaces to be stripped
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": " create ",
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Bad code
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": "abc/def",
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Empty name
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": TWD.code,
                                          "name": " "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Duplicated code
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": EUR.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], edit_uri)

        # Change code
        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": TWD.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_c_uri)

        response = self.client.get(detail_uri)
        self.assertEqual(response.status_code, 404)

        response = self.client.get(detail_c_uri)
        self.assertEqual(response.status_code, 200)

    def test_update_not_modified(self) -> None:
        """Tests that the data is not modified.

        :return: None.
        """
        from accounting.models import Currency
        detail_uri: str = f"{PREFIX}/{USD.code}"
        update_uri: str = f"{PREFIX}/{USD.code}/update"
        currency: Currency | None
        response: httpx.Response

        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": f" {USD.code} ",
                                          "name": f" {USD.name} "})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertIsNotNone(currency)
            currency.created_at \
                = currency.created_at - timedelta(seconds=5)
            currency.updated_at = currency.created_at
            db.session.commit()

        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": USD.code,
                                          "name": TWD.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertIsNotNone(currency)
            self.assertLess(currency.created_at,
                            currency.updated_at)

    def test_created_updated_by(self) -> None:
        """Tests the created-by and updated-by record.

        :return: None.
        """
        from accounting.models import Currency
        editor_username, admin_username = "editor", "admin"
        client, csrf_token = get_client(self.app, admin_username)
        detail_uri: str = f"{PREFIX}/{USD.code}"
        update_uri: str = f"{PREFIX}/{USD.code}/update"
        currency: Currency
        response: httpx.Response

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.created_by.username, editor_username)
            self.assertEqual(currency.updated_by.username, editor_username)

        response = client.post(update_uri,
                               data={"csrf_token": csrf_token,
                                     "code": USD.code,
                                     "name": f"{USD.name}-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.created_by.username, editor_username)
            self.assertEqual(currency.updated_by.username, admin_username)

    def test_api_exists(self) -> None:
        """Tests the API to check if a code exists.

        :return: None.
        """
        response: httpx.Response

        response = self.client.get(
            f"/accounting/api/currencies/exists-code?q={USD.code}")
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(set(data.keys()), {"exists"})
        self.assertTrue(data["exists"])

        response = self.client.get(
            f"/accounting/api/currencies/exists-code?q={USD.code}-1")
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(set(data.keys()), {"exists"})
        self.assertFalse(data["exists"])

    def test_l10n(self) -> None:
        """Tests the localization.

        :return: None
        """
        from accounting.models import Currency
        detail_uri: str = f"{PREFIX}/{USD.code}"
        update_uri: str = f"{PREFIX}/{USD.code}/update"
        currency: Currency
        response: httpx.Response

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.name_l10n, USD.name)
            self.assertEqual(currency.l10n, [])

        set_locale(self.client, self.csrf_token, "zh_Hant")

        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": USD.code,
                                          "name": f"{USD.name}-zh_Hant"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.name_l10n, USD.name)
            self.assertEqual({(x.locale, x.name) for x in currency.l10n},
                             {("zh_Hant", f"{USD.name}-zh_Hant")})

        set_locale(self.client, self.csrf_token, "en")

        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": USD.code,
                                          "name": f"{USD.name}-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.name_l10n, f"{USD.name}-2")
            self.assertEqual({(x.locale, x.name) for x in currency.l10n},
                             {("zh_Hant", f"{USD.name}-zh_Hant")})

        set_locale(self.client, self.csrf_token, "zh_Hant")

        response = self.client.post(update_uri,
                                    data={"csrf_token": self.csrf_token,
                                          "code": USD.code,
                                          "name": f"{USD.name}-zh_Hant-2"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        with self.app.app_context():
            currency = db.session.get(Currency, USD.code)
            self.assertEqual(currency.name_l10n, f"{USD.name}-2")
            self.assertEqual({(x.locale, x.name) for x in currency.l10n},
                             {("zh_Hant", f"{USD.name}-zh_Hant-2")})

    def test_delete(self) -> None:
        """Tests to delete a currency.

        :return: None.
        """
        from accounting.models import Currency
        detail_uri: str = f"{PREFIX}/{JPY.code}"
        delete_uri: str = f"{PREFIX}/{JPY.code}/delete"
        list_uri: str = PREFIX
        response: httpx.Response

        response = self.client.post(f"{PREFIX}/store",
                                    data={"csrf_token": self.csrf_token,
                                          "code": JPY.code,
                                          "name": JPY.name})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], detail_uri)

        add_journal_entry(self.client,
                          form={"csrf_token": self.csrf_token,
                                "next": NEXT_URI,
                                "date": date.today().isoformat(),
                                "currency-1-code": EUR.code,
                                "currency-1-credit-1-account_code": "1111-001",
                                "currency-1-credit-1-amount": "20"})

        with self.app.app_context():
            self.assertEqual({x.code for x in Currency.query.all()},
                             {USD.code, EUR.code, JPY.code})

        # Cannot delete the default currency
        response = self.client.post(f"{PREFIX}/{USD.code}/delete",
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{USD.code}")

        # Cannot delete the account that is in use
        response = self.client.post(f"{PREFIX}/{EUR.code}/delete",
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], f"{PREFIX}/{EUR.code}")

        # Success
        response = self.client.get(detail_uri)
        self.assertEqual(response.status_code, 200)
        response = self.client.post(delete_uri,
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], list_uri)

        with self.app.app_context():
            self.assertEqual({x.code for x in Currency.query.all()},
                             {USD.code, EUR.code})

        response = self.client.get(detail_uri)
        self.assertEqual(response.status_code, 404)
        response = self.client.post(delete_uri,
                                    data={"csrf_token": self.csrf_token})
        self.assertEqual(response.status_code, 404)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The report chooser.

This file is largely taken from the NanoParma ERP project, first written in
2021/9/16 by imacat (imacat@nanoparma.com).

"""
import re
import typing as t

from flask_babel import LazyString

from accounting import db
from accounting.locale import gettext
from accounting.models import Currency, Account
from accounting.report.period import Period, get_period
from accounting.template_globals import default_currency_code
from accounting.utils.current_account import CurrentAccount
from .option_link import OptionLink
from .report_type import ReportType
from .urls import journal_url, ledger_url, income_expenses_url, \
    trial_balance_url, income_statement_url, balance_sheet_url, unapplied_url


class ReportChooser:
    """The report chooser."""

    def __init__(self, active_report: ReportType,
                 period: Period | None = None,
                 currency: Currency | None = None,
                 account: Account | None = None):
        """Constructs the report chooser.

        :param active_report: The active report.
        :param period: The period.
        :param currency: The currency.
        :param account: The account.
        """
        self.__active_report: ReportType = active_report
        """The currently active report."""
        self.__period: Period = get_period() if period is None else period
        """The period."""
        self.__currency: Currency = db.session.get(
            Currency, default_currency_code()) \
            if currency is None else currency
        """The currency."""
        self.__account: Account = Account.cash() if account is None \
            else account
        """The currency."""
        self.__reports: list[OptionLink] = []
        """The links to the reports."""
        self.current_report: str | LazyString = ""
        """The title of the current report."""
        self.is_search: bool = active_report == ReportType.SEARCH
        """Whether the current report is the search page."""
        self.__reports.append(self.__income_expenses)
        self.__reports.append(self.__ledger)
        self.__reports.append(self.__journal)
        self.__reports.append(self.__trial_balance)
        self.__reports.append(self.__income_statement)
        self.__reports.append(self.__balance_sheet)
        self.__reports.append(self.__unapplied)
        for report in self.__reports:
            if report.is_active:
                self.current_report = report.title
        if self.is_search:
            self.current_report = gettext("Search")

    @property
    def __income_expenses(self) -> OptionLink:
        """Returns the income and expenses log.

        :return: The income and expenses log.
        """
        account: Account = self.__account
        if not re.match(r"[12][12]", account.base_code):
            account: Account = Account.cash()
        return OptionLink(gettext("Income and Expenses Log"),
                          income_expenses_url(self.__currency,
                                              CurrentAccount(account),
                                              self.__period),
                          self.__active_report == ReportType.INCOME_EXPENSES,
                          fa_icon="fa-solid fa-money-bill-wave")

    @property
    def __ledger(self) -> OptionLink:
        """Returns the ledger.

        :return: The ledger.
        """
        return OptionLink(gettext("Ledger"),
                          ledger_url(self.__currency, self.__account,
                                     self.__period),
                          self.__active_report == ReportType.LEDGER,
                          fa_icon="fa-solid fa-clipboard")

    @property
    def __journal(self) -> OptionLink:
        """Returns the journal.

        :return: The journal.
        """
        return OptionLink(gettext("Journal"), journal_url(self.__period),
                          self.__active_report == ReportType.JOURNAL,
                          fa_icon="fa-solid fa-book")

    @property
    def __trial_balance(self) -> OptionLink:
        """Returns the trial balance.

        :return: The trial balance.
        """
        return OptionLink(gettext("Trial Balance"),
                          trial_balance_url(self.__currency, self.__period),
                          self.__active_report == ReportType.TRIAL_BALANCE,
                          fa_icon="fa-solid fa-scale-unbalanced")

    @property
    def __income_statement(self) -> OptionLink:
        """Returns the income statement.

        :return: The income statement.
        """
        return OptionLink(gettext("Income Statement"),
                          income_statement_url(self.__currency, self.__period),
                          self.__active_report == ReportType.INCOME_STATEMENT,
                          fa_icon="fa-solid fa-file-invoice-dollar")

    @property
    def __balance_sheet(self) -> OptionLink:
        """Returns the balance sheet.

        :return: The balance sheet.
        """
        return OptionLink(gettext("Balance Sheet"),
                          balance_sheet_url(self.__currency, self.__period),
                          self.__active_report == ReportType.BALANCE_SHEET,
                          fa_icon="fa-solid fa-scale-balanced")

    @property
    def __unapplied(self) -> OptionLink:
        """Returns the unapplied original line items.

        :return: The unapplied original line items.
        """
        account: Account = self.__account
        if not account.is_need_offset:
            return OptionLink(gettext("Unapplied Original Line Items"),
                              unapplied_url(None),
                              self.__active_report == ReportType.UNAPPLIED,
                              fa_icon="fa-solid fa-link-slash")
        return OptionLink(gettext("Unapplied Original Line Items"),
                          unapplied_url(account),
                          self.__active_report == ReportType.UNAPPLIED,
                          fa_icon="fa-solid fa-link-slash")

    def __iter__(self) -> t.Iterator[OptionLink]:
        """Returns the iteration of the reports.

        :return: The iteration of the reports.
        """
        return iter(self.__reports)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/30

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The console commands for the account management.

"""
import typing as t
from secrets import randbelow

import click

from accounting import db
from accounting.models import BaseAccount, Account, AccountL10n
from accounting.utils.user import get_user_pk
import sqlalchemy as sa

AccountData = tuple[int, str, int, str, str, str, bool]
"""The format of the account data, as a list of (ID, base account code, number,
English, Traditional Chinese, Simplified Chinese, is-need-offset) tuples."""


def init_accounts_command(username: str) -> None:
    """Initializes the accounts."""
    creator_pk: int = get_user_pk(username)

    bases: list[BaseAccount] = BaseAccount.query\
        .filter(db.func.length(BaseAccount.code) == 4)\
        .order_by(BaseAccount.code).all()
    if len(bases) == 0:
        raise click.Abort

    existing: list[Account] = Account.query.all()

    existing_base_code: set[str] = {x.base_code for x in existing}
    bases_to_add: list[BaseAccount] = [x for x in bases
                                       if x.code not in existing_base_code]
    if len(bases_to_add) == 0:
        return

    existing_id: set[int] = {x.id for x in existing}

    def get_new_id() -> int:
        """Returns a new random account ID.

        :return: The newly-generated random account ID.
        """
        while True:
            new_id: int = 100000000 + randbelow(900000000)
            if new_id not in existing_id:
                existing_id.add(new_id)
                return new_id

    data: list[dict[str, t.Any]] = []
    l10n_data: list[dict[str, t.Any]] = []
    for base in bases_to_add:
        l10n: dict[str, str] = {x.locale: x.title for x in base.l10n}
        account_id: int = get_new_id()
        data.append({"id": account_id,
                     "base_code": base.code,
                     "no": 1,
                     "title_l10n": base.title_l10n,
                     "is_need_offset": __is_need_offset(base.code),
                     "created_by_id": creator_pk,
                     "updated_by_id": creator_pk})
        for locale in {"zh_Hant", "zh_Hans"}:
            l10n_data.append({"account_id": account_id,
                              "locale": locale,
                              "title": l10n[locale]})
    db.session.execute(sa.insert(Account), data)
    db.session.execute(sa.insert(AccountL10n), l10n_data)


def __is_need_offset(base_code: str) -> bool:
    """Checks that whether journal entry line items in the account need offset.

    :param base_code: The code of the base account.
    :return: True if journal entry line items in the account need offset, or
        False otherwise.
    """
    # Assets
    if base_code[0] == "1":
        if base_code[:3] in {"113", "114", "118", "184", "186"}:
            return True
        if base_code in {"1286", "1411", "1421", "1431", "1441", "1511",
                         "1521", "1581", "1611", "1851"}:
            return True
        return False
    # Liabilities
    if base_code[0] == "2":
        if base_code in {"2111", "2114", "2284", "2293", "2861"}:
            return False
        return True
    # Only assets and liabilities need offset
    return False



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/26

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the base account management.

"""
import unittest

import httpx
from flask import Flask

from testlib import create_test_app, get_client

LIST_URI: str = "/accounting/base-accounts"
"""The list URI."""
DETAIL_URI: str = "/accounting/base-accounts/1111"
"""The detail URI."""


class BaseAccountTestCase(unittest.TestCase):
    """The base account test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.app: Flask = create_test_app()

    def test_nobody(self) -> None:
        """Test the permission as nobody.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "nobody")
        response: httpx.Response

        response = client.get(LIST_URI)
        self.assertEqual(response.status_code, 403)

        response = client.get(DETAIL_URI)
        self.assertEqual(response.status_code, 403)

    def test_viewer(self) -> None:
        """Test the permission as viewer.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "viewer")
        response: httpx.Response

        response = client.get(LIST_URI)
        self.assertEqual(response.status_code, 200)

        response = client.get(DETAIL_URI)
        self.assertEqual(response.status_code, 200)

    def test_editor(self) -> None:
        """Test the permission as editor.

        :return: None.
        """
        client, csrf_token = get_client(self.app, "editor")
        response: httpx.Response

        response = client.get(LIST_URI)
        self.assertEqual(response.status_code, 200)

        response = client.get(DETAIL_URI)
        self.assertEqual(response.status_code, 200)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/22

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The getter and setter for the option management.

"""
import json

import sqlalchemy as sa

from accounting import db
from accounting.models import Option, Account, Currency
from accounting.utils.current_account import CurrentAccount
from accounting.utils.user import get_current_user_pk


class RecurringItem:
    """A recurring item."""

    def __init__(self, name: str, account_code: str,
                 description_template: str):
        """Constructs the recurring item.

        :param name: The name.
        :param account_code: The account code.
        :param description_template: The description template.
        """
        self.name: str = name
        self.account_code: str = account_code
        self.description_template: str = description_template

    @property
    def account_text(self) -> str:
        """Returns the account text.

        :return: The account text.
        """
        return str(Account.find_by_code(self.account_code))


class Recurring:
    """The recurring expenses or incomes."""

    def __init__(self, data: dict[str, list[tuple[str, str, str]]]):
        """Constructs the recurring item.

        :param data: The data.
        """
        self.expenses: list[RecurringItem] \
            = [RecurringItem(x[0], x[1], x[2]) for x in data["expense"]]
        self.incomes: list[RecurringItem] \
            = [RecurringItem(x[0], x[1], x[2]) for x in data["income"]]

    @property
    def codes(self) -> set[str]:
        """Returns all the account codes.

        :return: All the account codes.
        """
        return {x.account_code for x in self.expenses + self.incomes}


class Options:
    """The options."""

    def __init__(self):
        """Constructs the options."""
        self.is_modified: bool = False
        """Whether the options were modified."""

    @property
    def default_currency_code(self) -> str:
        """Returns the default currency code.

        :return: The default currency code.
        """
        return self.__get_option("default_currency_code", "USD")

    @default_currency_code.setter
    def default_currency_code(self, value: str) -> None:
        """Sets the default currency code.

        :param value: The default currency code.
        :return: None.
        """
        self.__set_option("default_currency_code", value)

    @property
    def default_currency(self) -> Currency:
        """Returns the default currency.

        :return: The default currency.
        """
        return db.session.get(Currency, self.default_currency_code)

    @property
    def default_ie_account_code(self) -> str:
        """Returns the default account code for the income and expenses log.

        :return: The default account code for the income and expenses log.
        """
        return self.__get_option("default_ie_account", Account.CASH_CODE)

    @default_ie_account_code.setter
    def default_ie_account_code(self, value: str) -> None:
        """Sets the default account code for the income and expenses log.

        :param value: The default account code for the income and expenses log.
        :return: None.
        """
        self.__set_option("default_ie_account", value)

    @property
    def default_ie_account(self) -> CurrentAccount:
        """Returns the default account for the income and expenses log.

        :return: The default account for the income and expenses log.
        """
        if self.default_ie_account_code \
                == CurrentAccount.CURRENT_AL_CODE:
            return CurrentAccount.current_assets_and_liabilities()
        return CurrentAccount(
            Account.find_by_code(self.default_ie_account_code))

    @property
    def recurring_data(self) -> dict[str, list[tuple[str, str, str]]]:
        """Returns the data of the recurring expenses and incomes.

        :return: The data of the recurring expenses and incomes.
        """
        json_data: str | None = self.__get_option("recurring")
        if json_data is None:
            return {"expense": [], "income": []}
        return json.loads(json_data)

    @recurring_data.setter
    def recurring_data(self,
                       value: dict[str, list[tuple[str, str, str]]]) -> None:
        """Sets the data of the recurring expenses and incomes.

        :param value: The data of the recurring expenses and incomes.
        :return: None.
        """
        self.__set_option("recurring", json.dumps(value, ensure_ascii=False,
                                                  separators=(",", ":")))

    @property
    def recurring(self) -> Recurring:
        """Returns the recurring expenses and incomes.

        :return: The recurring expenses and incomes.
        """
        return Recurring(self.recurring_data)

    @staticmethod
    def __get_option(name: str, default: str | None = None) -> str:
        """Returns the value of an option.

        :param name: The name.
        :param default: The default value when the value does not exist.
        :return: The value.
        """
        option: Option | None = db.session.get(Option, name)
        if option is None:
            return default
        return option.value

    def __set_option(self, name: str, value: str) -> None:
        """Sets the value of an option.

        :param name: The name.
        :param value: The value.
        :return: None.
        """
        option: Option | None = db.session.get(Option, name)
        if option is None:
            current_user_pk: int = get_current_user_pk()
            db.session.add(Option(name=name,
                                  value=value,
                                  created_by_id=current_user_pk,
                                  updated_by_id=current_user_pk))
            self.is_modified = True
            return
        if option.value == value:
            return
        option.value = value
        option.updated_by_id = get_current_user_pk()
        option.updated_at = sa.func.now()
        self.is_modified = True

    def commit(self) -> None:
        """Commits the options to the database.

        :return: None.
        """
        db.session.commit()
        self.is_modified = False


options: Options = Options()
"""The options."""



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/26

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The views for the base account management.

"""
from flask import Blueprint, render_template

from accounting.models import BaseAccount
from accounting.utils.pagination import Pagination
from accounting.utils.permission import has_permission, can_view
from .queries import get_base_account_query

bp: Blueprint = Blueprint("base-account", __name__)
"""The view blueprint for the base account management."""


@bp.get("", endpoint="list")
@has_permission(can_view)
def list_accounts() -> str:
    """Lists the base accounts.

    :return: The account list.
    """
    accounts: list[BaseAccount] = get_base_account_query()
    pagination: Pagination = Pagination[BaseAccount](accounts)
    return render_template("accounting/base-account/list.html",
                           list=pagination.list, pagination=pagination)


@bp.get("<baseAccount:account>", endpoint="detail")
@has_permission(can_view)
def show_account_detail(account: BaseAccount) -> str:
    """Shows the account detail.

    :param account: The account.
    :return: The detail.
    """
    return render_template("accounting/base-account/detail.html", obj=account)




# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/27

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The common test libraries for the journal entry test cases.

"""
import re
from datetime import date
from decimal import Decimal
from secrets import randbelow

from flask import Flask

from test_site import db
from testlib import NEXT_URI, Accounts

NON_EMPTY_NOTE: str = "  This is \n\na test."
"""The stripped content of an non-empty note."""
EMPTY_NOTE: str = " \n\n  "
"""The empty note content."""


def get_add_form(csrf_token: str) -> dict[str, str]:
    """Returns the form data to add a new journal entry.

    :param csrf_token: The CSRF token.
    :return: The form data to add a new journal entry.
    """
    return {"csrf_token": csrf_token,
            "next": NEXT_URI,
            "date": date.today().isoformat(),
            "currency-0-code": "USD",
            "currency-0-debit-0-no": "16",
            "currency-0-debit-0-account_code": Accounts.CASH,
            "currency-0-debit-0-description": " ",
            "currency-0-debit-0-amount": " 495.26 ",
            "currency-0-debit-6-no": "2",
            "currency-0-debit-6-account_code": Accounts.BANK,
            "currency-0-debit-6-description": " Deposit ",
            "currency-0-debit-6-amount": "6000",
            "currency-0-debit-12-no": "2",
            "currency-0-debit-12-account_code": Accounts.OFFICE,
            "currency-0-debit-12-description": " Pens ",
            "currency-0-debit-12-amount": "4.99",
            "currency-0-credit-2-no": "6",
            "currency-0-credit-2-account_code": Accounts.SERVICE,
            "currency-0-credit-2-description": " ",
            "currency-0-credit-2-amount": "5500",
            "currency-0-credit-7-account_code": Accounts.SALES,
            "currency-0-credit-7-description": " ",
            "currency-0-credit-7-amount": "950",
            "currency-0-credit-27-account_code": Accounts.INTEREST,
            "currency-0-credit-27-description": " ",
            "currency-0-credit-27-amount": "50.25",
            "currency-3-no": "2",
            "currency-3-code": "JPY",
            "currency-3-debit-2-no": "2",
            "currency-3-debit-2-account_code": Accounts.CASH,
            "currency-3-debit-2-description": " ",
            "currency-3-debit-2-amount": "15000",
            "currency-3-debit-9-no": "5",
            "currency-3-debit-9-account_code": Accounts.BANK,
            "currency-3-debit-9-description": " Deposit ",
            "currency-3-debit-9-amount": "95000",
            "currency-3-credit-3-account_code": Accounts.AGENCY,
            "currency-3-credit-3-description": " Realtor ",
            "currency-3-credit-3-amount": "65000",
            "currency-3-credit-5-no": "4",
            "currency-3-credit-5-account_code": Accounts.DONATION,
            "currency-3-credit-5-description": " Donation ",
            "currency-3-credit-5-amount": "45000",
            "currency-16-code": "TWD",
            "currency-16-debit-2-no": "2",
            "currency-16-debit-2-account_code": Accounts.CASH,
            "currency-16-debit-2-description": " ",
            "currency-16-debit-2-amount": "10000",
            "currency-16-debit-9-no": "2",
            "currency-16-debit-9-account_code": Accounts.TRAVEL,
            "currency-16-debit-9-description": " Gas ",
            "currency-16-debit-9-amount": "30000",
            "currency-16-credit-6-no": "6",
            "currency-16-credit-6-account_code": Accounts.RENT_INCOME,
            "currency-16-credit-6-description": " Rent ",
            "currency-16-credit-6-amount": "35000",
            "currency-16-credit-9-account_code": Accounts.DONATION,
            "currency-16-credit-9-description": " Donation ",
            "currency-16-credit-9-amount": "5000",
            "note": f"\n \n\n  \n{NON_EMPTY_NOTE}  \n  \n\n  "}


def get_unchanged_update_form(journal_entry_id: int, app: Flask,
                              csrf_token: str) -> dict[str, str]:
    """Returns the form data to update a journal entry, where the data are not
    changed.

    :param journal_entry_id: The journal entry ID.
    :param app: The Flask application.
    :param csrf_token: The CSRF token.
    :return: The form data to update the journal entry, where the data are not
        changed.
    """
    from accounting.models import JournalEntry, JournalEntryCurrency
    with app.app_context():
        journal_entry: JournalEntry | None \
            = db.session.get(JournalEntry, journal_entry_id)
        assert journal_entry is not None
        currencies: list[JournalEntryCurrency] = journal_entry.currencies

    form: dict[str, str] \
        = {"csrf_token": csrf_token,
           "next": NEXT_URI,
           "date": journal_entry.date,
           "note": "  \n \n\n  " if journal_entry.note is None
           else f"\n    \n\n \n  \n{journal_entry.note}  \n\n   "}
    currency_indices_used: set[int] = set()
    currency_no: int = 0
    for currency in currencies:
        currency_index: int = __get_new_index(currency_indices_used)
        currency_no = currency_no + 3 + randbelow(3)
        currency_prefix: str = f"currency-{currency_index}"
        form[f"{currency_prefix}-no"] = str(currency_no)
        form[f"{currency_prefix}-code"] = currency.code
        line_item_indices_used: set[int]
        line_item_no: int
        prefix: str

        line_item_indices_used = set()
        line_item_no = 0
        for line_item in currency.debit:
            line_item_index: int = __get_new_index(line_item_indices_used)
            line_item_no = line_item_no + 3 + randbelow(3)
            prefix = f"{currency_prefix}-debit-{line_item_index}"
            form[f"{prefix}-id"] = str(line_item.id)
            form[f"{prefix}-no"] = str(line_item_no)
            form[f"{prefix}-account_code"] = line_item.account.code
            form[f"{prefix}-description"] \
                = "  " if line_item.description is None \
                else f" {line_item.description} "
            form[f"{prefix}-amount"] = str(line_item.amount)

        line_item_indices_used = set()
        line_item_no = 0
        for line_item in currency.credit:
            line_item_index: int = __get_new_index(line_item_indices_used)
            line_item_no = line_item_no + 3 + randbelow(3)
            prefix = f"{currency_prefix}-credit-{line_item_index}"
            form[f"{prefix}-id"] = str(line_item.id)
            form[f"{prefix}-no"] = str(line_item_no)
            form[f"{prefix}-account_code"] = line_item.account.code
            form[f"{prefix}-description"] \
                = "  " if line_item.description is None \
                else f" {line_item.description} "
            form[f"{prefix}-amount"] = str(line_item.amount)

    return form


def __get_new_index(indices_used: set[int]) -> int:
    """Returns a new random index that is not used.

    :param indices_used: The set of indices that are already used.
    :return: The newly-generated random index that is not used.
    """
    while True:
        index: int = randbelow(100)
        if index not in indices_used:
            indices_used.add(index)
            return index


def get_update_form(journal_entry_id: int, app: Flask,
                    csrf_token: str, is_debit: bool | None) -> dict[str, str]:
    """Returns the form data to update a journal entry, where the data are
    changed.

    :param journal_entry_id: The journal entry ID.
    :param app: The Flask application.
    :param csrf_token: The CSRF token.
    :param is_debit: True for a cash disbursement journal entry, False for a
        cash receipt journal entry, or None for a transfer journal entry.
    :return: The form data to update the journal entry, where the data are
        changed.
    """
    form: dict[str, str] = get_unchanged_update_form(
        journal_entry_id, app, csrf_token)

    # Mess up the line items in a currency
    currency_prefix: str = __get_currency_prefix(form, "USD")
    if is_debit is None or is_debit:
        form = __mess_up_debit(form, currency_prefix)
    if is_debit is None or not is_debit:
        form = __mess_up_credit(form, currency_prefix)

    # Mess-up the currencies
    form = __mess_up_currencies(form)

    return form


def __mess_up_debit(form: dict[str, str], currency_prefix: str) \
        -> dict[str, str]:
    """Mess up the debit line items in the form data.

    :param form: The form data.
    :param currency_prefix: The key prefix of the currency sub-form.
    :return: The messed-up form.
    """
    key: str
    m: re.Match

    # Remove the office disbursement
    key = [x for x in form
           if x.startswith(currency_prefix)
           and form[x] == Accounts.OFFICE][0]
    m = re.match(r"^((.+-)\d+-)account_code$", key)
    debit_prefix: str = m.group(2)
    line_item_prefix: str = m.group(1)
    amount: Decimal = Decimal(form[f"{line_item_prefix}amount"])
    form = {x: form[x] for x in form if not x.startswith(line_item_prefix)}
    # Add a new travel disbursement
    indices: set[int] = set()
    for key in form:
        m = re.match(r"^.+-(\d+)-amount$", key)
        if m is not None:
            indices.add(int(m.group(1)))
    new_index: int = max(indices) + 5 + randbelow(20)
    min_no: int = min({int(form[x]) for x in form if x.endswith("-no")
                       and x.startswith(debit_prefix)})
    form[f"{debit_prefix}{new_index}-no"] = str(1 + randbelow(min_no - 1))
    form[f"{debit_prefix}{new_index}-amount"] = str(amount)
    form[f"{debit_prefix}{new_index}-account_code"] = Accounts.TRAVEL
    # Swap the cash and the bank order
    key_cash: str = __get_line_item_no_key(
        form, currency_prefix, Accounts.CASH)
    key_bank: str = __get_line_item_no_key(
        form, currency_prefix, Accounts.BANK)
    form[key_cash], form[key_bank] = form[key_bank], form[key_cash]
    return form


def __mess_up_credit(form: dict[str, str], currency_prefix: str) \
        -> dict[str, str]:
    """Mess up the credit line items in the form data.

    :param form: The form data.
    :param currency_prefix: The key prefix of the currency sub-form.
    :return: The messed-up form.
    """
    key: str
    m: re.Match

    # Remove the sales receipt
    key = [x for x in form
           if x.startswith(currency_prefix)
           and form[x] == Accounts.SALES][0]
    m = re.match(r"^((.+-)\d+-)account_code$", key)
    credit_prefix: str = m.group(2)
    line_item_prefix: str = m.group(1)
    amount: Decimal = Decimal(form[f"{line_item_prefix}amount"])
    form = {x: form[x] for x in form if not x.startswith(line_item_prefix)}
    # Add a new agency receipt
    indices: set[int] = set()
    for key in form:
        m = re.match(r"^.+-(\d+)-amount$", key)
        if m is not None:
            indices.add(int(m.group(1)))
    new_index: int = max(indices) + 5 + randbelow(20)
    min_no: int = min({int(form[x]) for x in form if x.endswith("-no")
                       and x.startswith(credit_prefix)})
    form[f"{credit_prefix}{new_index}-no"] = str(1 + randbelow(min_no - 1))
    form[f"{credit_prefix}{new_index}-amount"] = str(amount)
    form[f"{credit_prefix}{new_index}-account_code"] = Accounts.AGENCY
    # Swap the service and the interest order
    key_srv: str = __get_line_item_no_key(
        form, currency_prefix, Accounts.SERVICE)
    key_int: str = __get_line_item_no_key(
        form, currency_prefix, Accounts.INTEREST)
    form[key_srv], form[key_int] = form[key_int], form[key_srv]
    return form


def __mess_up_currencies(form: dict[str, str]) -> dict[str, str]:
    """Mess up the currency sub-forms in the form data.

    :param form: The form data.
    :return: The messed-up form.
    """
    key: str
    m: re.Match

    # Remove JPY
    currency_prefix: str = __get_currency_prefix(form, "JPY")
    form = {x: form[x] for x in form if not x.startswith(currency_prefix)}
    # Add AUD
    indices: set[int] = set()
    for key in form:
        m = re.match(r"^currency-(\d+)-code$", key)
        if m is not None:
            indices.add(int(m.group(1)))
    new_index: int = max(indices) + 5 + randbelow(20)
    min_no: int = min({int(form[x]) for x in form if x.endswith("-no")
                       and "-debit-" not in x and "-credit-" not in x})
    prefix: str = f"currency-{new_index}-"
    form.update({
        f"{prefix}code": "AUD",
        f"{prefix}no": str(1 + randbelow(min_no - 1)),
        f"{prefix}debit-0-no": "6",
        f"{prefix}debit-0-account_code": Accounts.OFFICE,
        f"{prefix}debit-0-description": " Envelop ",
        f"{prefix}debit-0-amount": "5.45",
        f"{prefix}debit-14-no": "6",
        f"{prefix}debit-14-account_code": Accounts.CASH,
        f"{prefix}debit-14-description": "  ",
        f"{prefix}debit-14-amount": "14.55",
        f"{prefix}credit-16-no": "7",
        f"{prefix}credit-16-account_code": Accounts.RENT_INCOME,
        f"{prefix}credit-16-description": " Bike ",
        f"{prefix}credit-16-amount": "19.5",
        f"{prefix}credit-22-no": "5",
        f"{prefix}credit-22-account_code": Accounts.DONATION,
        f"{prefix}credit-22-description": " Artist ",
        f"{prefix}credit-22-amount": "0.5",
    })
    # Swap the USD and TWD order
    usd_prefix: str = __get_currency_prefix(form, "USD")
    key_usd: str = f"{usd_prefix}no"
    twd_prefix: str = __get_currency_prefix(form, "TWD")
    key_twd: str = f"{twd_prefix}no"
    form[key_usd], form[key_twd] = form[key_twd], form[key_usd]
    # Change TWD to EUR
    key = [x for x in form if form[x] == "TWD"][0]
    form[key] = "EUR"
    return form


def __get_line_item_no_key(form: dict[str, str], currency_prefix: str,
                           code: str) -> str:
    """Returns the key of a line item number in the form data.

    :param form: The form data.
    :param currency_prefix: The prefix of the currency.
    :param code: The code of the account.
    :return: The key of the line item number in the form data.
    """
    key: str = [x for x in form
                if x.startswith(currency_prefix)
                and form[x] == code][0]
    m: re.Match = re.match(r"^(.+-\d+-)account_code$", key)
    return f"{m.group(1)}no"


def __get_currency_prefix(form: dict[str, str], code: str) -> str:
    """Returns the prefix of a currency in the form data.

    :param form: The form data.
    :param code: The code of the currency.
    :return: The prefix of the currency.
    """
    key: str = [x for x in form if form[x] == code][0]
    m: re.Match = re.match(r"^(.+-)code$", key)
    return m.group(1)


def set_negative_amount(form: dict[str, str]) -> None:
    """Sets a negative amount in the form data, keeping the balance.

    :param form: The form data.
    :return: None.
    """
    amount_keys: list[str] = []
    prefix: str = ""
    for key in form.keys():
        m: re.Match = re.match(r"^(.+)-\d+-amount$", key)
        if m is None:
            continue
        if prefix != "" and prefix != m.group(1):
            continue
        prefix = m.group(1)
        amount_keys.append(key)
    form[amount_keys[0]] = str(-Decimal(form[amount_keys[0]]))
    form[amount_keys[1]] = str(Decimal(form[amount_keys[1]])
                               + 2 * Decimal(form[amount_keys[0]]))


def remove_debit_in_a_currency(form: dict[str, str]) -> None:
    """Removes debit line items in a currency sub-form.

    :param form: The form data.
    :return: None.
    """
    key: str = [x for x in form if "-debit-" in x][0]
    m: re.Match = re.match(r"^(.+-debit-)", key)
    keys: set[str] = {x for x in form if x.startswith(m.group(1))}
    for key in keys:
        del form[key]


def remove_credit_in_a_currency(form: dict[str, str]) -> None:
    """Removes credit line items in a currency sub-form.

    :param form: The form data.
    :return: None.
    """
    key: str = [x for x in form if "-credit-" in x][0]
    m: re.Match = re.match(r"^(.+-credit-)", key)
    keys: set[str] = {x for x in form if x.startswith(m.group(1))}
    for key in keys:
        del form[key]
